import getopt, sys
from .ui_lib import print_handles

def show_handle(client, rest):
        opts, args = getopt.getopt(rest, "j")
        opts = dict(opts)
        if len(args) < 1:
            print(Usage)
            sys.exit(2)
        did = args[0]
        handle = client.get_handle(handle_id)
        if not handle:
            print(f"Handle {handle_id} not found")
            sys.exit(1)
        if "-j" in opts:
            print(pretty_json(handle))
        else:
            for name, val in handle.items():
                if name != "replicas":
                    print(f"{name:10s}: {val}")
            if "replicas" in handle:
                print("replicas:")
                replicas = handle["replicas"] or []
                replicas = sorted(replicas, key=lambda r: (-r["preference"], r["rse"]))
                for r in handle["replicas"]:
                    print("  Preference: ", r["preference"])
                    print("  RSE:        ", r["rse"])
                    print("  Path:       ", r["path"] or "")
                    print("  URL:        ", r["url"] or "")
                    print("  Available:  ", "yes" if r["available"] else "no")

def list_handles(client, rest):
        opts, args = getopt.getopt(rest, "p:s:r:")
        opts = dict(opts)
        project_id = opts.get("-p")
        if project_id is not None:
            project_id = int(project_id)
        lst = client.list_handles(project_id=project_id, rse=opts.get("-r"), status=opts.get("-s"))
        print_handles(lst)



import json, time, io
from datetime import datetime, timedelta, timezone
from metacat.auth import BaseDBUser as DBUser

def cursor_iterator(c):
    t = c.fetchone()
    while t is not None:
        yield t
        t = c.fetchone()


def json_literal(v):
    if isinstance(v, str):       v = '"%s"' % (v,)
    elif isinstance(v, bool):    v = "true" if v else "false"
    elif v is None:              v = "null"
    else:   v = str(v)
    return v

class DBObject(object):
    
    @classmethod
    def from_tuple(cls, db, dbtup):
        h = cls(db, *dbtup)
        return h
    
    @classmethod
    def columns(cls, table_name=None, as_text=True, exclude=[]):
        if isinstance(exclude, str):
            exclude = [exclude]
        clist = [c for c in cls.Columns if c not in exclude]
        if table_name:
            clist = [table_name+"."+cn for cn in clist]
        if as_text:
            return ",".join(clist)
        else:
            return clist
    
    @classmethod
    def pk_columns(cls, table_name=None, as_text=True, exclude=[]):
        if isinstance(exclude, str):
            exclude = [exclude]
        clist = [c for c in cls.PK if c not in exclude]
        if table_name:
            clist = [table_name+"."+cn for cn in clist]
        if as_text:
            return ",".join(clist)
        else:
            return clist
    
    @classmethod
    def get(cls, db, *pk_vals):
        pk_cols_values = [f"{c} = %s" for c in cls.PK]
        where = " and ".join(pk_cols_values)
        cols = ",".join(cls.Columns)
        c = db.cursor()
        c.execute(f"select {cols} from {cls.Table} where {where}", pk_vals)
        tup = c.fetchone()
        if tup is None: return None
        else:   return cls.from_tuple(db, tup)

    def _delete(self, cursor=None, do_commit=True, **pk_values):
        cursor = cursor or self.DB.cursor()
        where_clause = " and ".join(f"{column} = '{value}'" for column, value in pk_values.items())
        try:
            cursor.execute(f"""
                delete from {self.Table} where {where_clause}
            """)
            if do_commit:
                cursor.execute("commit")
        except:
            cursor.execute("rollback")
            raise
    
    @classmethod
    def list(db, cls):
        columns = cls.columns(as_text=True)
        table = cls.Table
        c = db.cursor()
        c.execute(f"select {columns} from {table}")
        return (cls.from_tuple(db, tup) for tup in cursor_iterator(c))
    
    def delete(self, cursor=None, do_commit=True, **pk_values):
        return self._delete(cursor=None, do_commit=True, **pk_values)
    
class DBManyToMany(object):
    
    def __init__(self, db, table, src_fk_values, dst_fk_columns, payload_columns, dst_class):
        self.DB = db
        self.Table = table
        self.SrcFKColumns, self.SrcFKValues = zip(*list(src_fk_values.items()))
        self.DstFKColumns = dst_fk_columns
        self.DstClass = dst_class
        self.DstTable = dst_class.Table
        self.DstPKColumns = self.DstTable.PK

    def add(self, dst_pk_values, payload, cursor=None, do_commit=True):
        assert len(dst_pk_values) == len(self.DstFKColumns)
        
        payload_cols_vals = list(payload.items())
        payload_cols, payload_vals = zip(*payload_cols_vals)
        
        fk_cols = ",".join(self.SrcFKColumns + self.DstFKColumns)
        cols = ",".join(self.SrcFKColumns + self.DstFKColumns + payload_cols)
        vals = ",".join([f"'{v}'" for v in self.SrcFKValues + dst_pk_values + payload_vals])
        
        if cursor is None: cursor = self.DB.cursor()
        try:
            cursor.execute(f"""
                insert into {self.Table}({cols}) values({vals})
                    on conflict({fk_cols}) do nothing
            """)
            if do_commit:
                cursor.execute("commit")
        except:
            cursor.execute(rollback)
            raise
        return self

    def list(self, cursor=None):
        out_columns = ",".join(f"{self.DstTable}.{c}" for c in self.DstClass.Columns)
        join_column_pairs = [
            (f"{self.Table}.{dst_fk}", f"{self.DstTable}.{dst_pk}") 
            for src_fk, dst_pk in zip(self.DstFKColumns, self.DstPKColumns)
        ]
        join_condition = " and ".join(f"{fk} = {pk}" for fk, pk in join_column_pairs)
        if cursor is None: cursor = self.DB.cursor()
        cursor.execute(f"""
            select {out_columns}
                from {self.DstTable}, {self.Table}
                where {join_condition}
        """)
        return (self.DstClass.from_tuple(self.DB, tup) for tup in fetch_generator(cursor))
        
    def __iter__(self):
        return self.list()

class DBOneToMany(object):
    
    def __init__(self, db, table, src_pk_values, dst_fk_columns, dst_class):
        self.DB = db
        self.Table = table
        self.SrcPKColumns, self.SrcPKValues = zip(*list(src_pk_values.items()))
        self.DstClass = dst_class
        self.DstTable = dst_class.Table
        self.DstFKColumns = dst_fk_columns

    def list(self, cursor=None):
        out_columns = ",".join(f"{self.DstTable}.{c}" for c in self.DstClass.Columns)
        join_column_pairs = [
            (f"{self.Table}.{dst_pk}", f"{self.DstTable}.{dst_fk}") 
            for src_pk, dst_fk in zip(self.SrcPKColumns, self.DstFKColumns)
        ]
        join_condition = " and ".join(f"{pk} = {fk}" for pk, fk in join_column_pairs)
        if cursor is None: cursor = self.DB.cursor()
        cursor.execute(f"""
            select {out_columns}
                from {self.DstTable}, {self.Table}
                where {join_condition}
        """)
        return (self.DstClass.from_tuple(self.DB, tup) for tup in fetch_generator(cursor))
        
    def __iter__(self):
        return self.list()

class HasLogRecord(object):

    def __init__(self, log_table, id_column):
        self.LogTable = log_table
        self.IDColumn = id_column
    
    class DBLogRecord(object):
    
        def __init__(self, type, t, message):
            self.Type = type
            self.T = t
            self.Message = message

    def id(self):
        raise NotImplementedError()

    def add_log(self, type, message):
        c = self.DB.cursor()
        c.execute(f"""
            begin;
            
            insert into {self.LogTable}({self.IDColumn}, type, message)
            values(%s, %s, %s);

            commit
        """, (self.ID, type, message))

    def log(self, type=None, since=None, reversed=False):
        my_id = self.id()
        wheres = [f"column={my_id}"]
        if isinstance(since, (float, int)):
            since = datetime.utcfromtimestamp(since).replace(tzinfo=timezone.utc)
            wheres.append(f"t >= {since}")
        if type is not None:
            wheres.append(f"type = '{type}'")
        wheres = "" if not wheres else " and " + " and ".join(wheres)
        desc = "desc" if reversed else ""
        sql = f"""
            select type, t, message from {self.LogTable}
                {wheres}
                order by t {desc}
        """
        c = self.DB.cursor()
        c.execute(sql)
        return (self.DBLogRecord(type, t, message) for type, t, message in cursor_iterator())
        

class DBProject(DBObject, HasLogRecord):
    
    InitialState = "active"
    
    Columns = "id,owner,created_timestamp,state,retry_count,attributes".split(",")
    Table = "projects"
    PK = ["id"]
    
    def __init__(self, db, id, owner=None, created_timestamp=None, state=None, retry_count=0, attributes={}):
        HasLogRecord.__init__(self, "project_log", "project_id")
        self.DB = db
        self.ID = id
        self.Owner = owner
        self.State = state
        self.CreatedTimestamp = created_timestamp               # datetime
        self.RetryCount = retry_count
        self.Attributes = attributes.copy()
        self.Handles = None
        self.HandleCounts = None
        
    def id(self):
        return str(self.ID)
        
    def as_jsonable(self, with_handles=False, with_replicas=False):
        #print("Project.as_jsonable: with_handles:", with_handles, "   with_replicas:", with_replicas)
        out = dict(
            project_id = self.ID,
            owner = self.Owner,
            state = self.State,
            retry_count = self.RetryCount,
            attributes = self.Attributes,
            created_timestamp = self.CreatedTimestamp.timestamp(),
            active = self.is_active()
        )
        if with_handles:
            out["file_handles"] = [h.as_jsonable(with_replicas=with_replicas) for h in self.handles()]
            #print("Project.as_jsonable: handles:", out["file_handles"])
        return out

    def attributes_as_json(self):
        return json.dumps(self.Attributes, indent=4)
        
    @staticmethod
    def create(db, owner, retry_count=None, attributes={}):
        if isinstance(owner, DBUser):
            owner = owner.Username
        c = db.cursor()
        try:
            c.execute("begin")
            c.execute("""
                insert into projects(owner, state, retry_count, attributes)
                    values(%s, %s, %s, %s)
                    returning id
            """, (owner, DBProject.InitialState, retry_count, json.dumps(attributes)))
            id = c.fetchone()[0]
            db.commit()
        except:
            db.rollback()
            raise
            
        return DBProject.get(db, id)

    @staticmethod
    def list(db, owner=None, state=None, not_state=None, attributes=None, with_handle_counts=False):
        wheres = ["true"]
        if owner: wheres.append(f"p.owner='{owner}'")
        if state: wheres.append(f"p.state='{state}'")
        if not_state: wheres.append(f"p.state!='{not_state}'")
        if attributes is not None:
            for name, value in attributes.items():
                wheres.append("p.attributes @> '{\"%s\": %s}'::jsonb" % (name, json_literal(value)))
        wheres = " and ".join(wheres)
        c = db.cursor()
        table = DBProject.Table
        columns = DBProject.columns("p", as_text=True)
        if with_handle_counts:
            h_table = DBFileHandle.Table
            c.execute(f"""
                select {columns}, h.state, count(*)
                    from {table} p, {h_table} h
                    where {wheres}
                        and p.id = h.project_id
                    group by {columns}, h.state
                    order by p.id
            """)
            p = None
            for tup in cursor_iterator(c):
                #print(tup)
                p_tuple, h_state, count = tup[:len(DBProject.Columns)], tup[-2], tup[-1]
                p1 = DBProject.from_tuple(db, p_tuple)
                if p is None or p.ID != p1.ID:
                    if p is not None:
                        yield p
                    p = p1
                    p.HandleCounts = {state:0 for state in DBFileHandle.States}
                p.HandleCounts[h_state] = count
            if p is not None:
                yield p
        else:
            c.execute(f"""
                select {columns}
                    from {table} p
                    where {wheres}
            """)
            for tup in cursor_iterator(c):
                yield DBProject.from_tuple(db, tup)
        
    def save(self):
        c = self.DB.cursor()
        try:
            c.execute("begin")
            c.execute("""
                update projects set state=%s
                    where id=%s
            """, (self.State, self.ID))
            self.DB.commit()
        except:
            self.DB.rollback()
            raise
        
    def handles(self, with_replicas=True):
        if self.Handles is None:
            self.Handles = list(DBFileHandle.list(self.DB, project_id=self.ID, with_replicas=with_replicas))
        return self.Handles
        
    def handle(self, namespace=None, name=None):
        return DBFileHandle.lookup(self.DB, project_id=self.ID, namespace=namespace, name=name)
            
    def add_files(self, files_descs):
        # files_descs is list of disctionaries: [{"namespace":..., "name":...}, ...]
        files_descs = list(files_descs)     # make sure it's not a generator
        DBFile.create_many(self.DB, files_descs)
        DBFileHandle.create_many(self.DB, self.ID, files_descs)
        
    def files(self):
        return DBFile.list(self.DB, self.ID)
        
    def reserve_next_file(self, worker_id):
        handle = DBFileHandle.reserve_next_available(self.DB, self.ID, worker_id)
        if handle is not None:
            did = handle.did()
            self.add_log("workflow", f"file {did} reserved for {worker_id}")
        return handle

    def is_active(self):
        lst = DBFileHandle.list(self.DB, project_id=self.ID, state=["ready", "reserved"])
        for h in lst:       # lst may be a generator
            return True
        else:
            return False
            
    def file_state_counts(self):
        counts = {}
        for h in DBFileHandle.list(self.DB, project_id=self.ID):
            s = h.State
            counts[s] = counts.get(s, 0) + 1
        return out
        
class DBFile(DBObject):
    
    Columns = ["namespace", "name"]
    PK = ["namespace", "name"]
    Table = "files"
    
    def __init__(self, db, namespace, name):
        self.DB = db
        self.Namespace = namespace
        self.Name = name
        self.Replicas = None	# {rse -> DBReplica}
    
    def id(self):
        return f"{self.Namespace}:{self.Name}"

    def did(self):
        return f"{self.Namespace}:{self.Name}"

    def as_jsonable(self, with_replicas=False):
        out = dict(
            namespace   = self.Namespace,
            name        = self.Name,
        )
        if with_replicas:
            out["replicas"] = {rse: r.as_jsonable() for rse, r in self.replicas().items()}
        return out
        
    def replicas(self):
        if self.Replicas is None:
            self.Replicas = {r.RSE: r for r in DBReplica.list(self.DB, self.Namespace, self.Name)}
        return self.Replicas

    def create_replica(self, rse, path, url, preference=0, available=False):
        DBReplica.create(self.DB, self.Namespace, self.Name, rse, path, url, preference=preference, available=available)
        self.Replicas = None	# force re-load from the DB
        
    def get_replica(self, rse):
        return self.replicas().get(rse)
        
    @staticmethod
    def create(db, namespace, name, error_if_exists=False):
        c = self.DB.cursor()
        try:
            c.execute("begin")
            conflict = "on conflict (namespace, name) do nothing" if not error_if_exists else ""
            c.execute(f"insert into files(namespace, name) values(%s, %s) {conflict}; commit" % (namespace, name))
            return DBFile.get(db, namespace, name)
        except:
            c.execute("rollback")
            raise

    @staticmethod
    def create_many(db, descs):
        #
        # descs: [{"namespace":..., "name":..., ...}]
        #
        csv = [f"%s\t%s" % (item["namespace"], item["name"]) for item in descs]
        data = io.StringIO("\n".join(csv))
        table = DBFile.Table
        c = db.cursor()
        try:
            t = int(time.time()*1000)
            temp_table = f"files_temp_{t}"
            c.execute("begin")
            c.execute(f"create temp table {temp_table} (namespace text, name text)")
            c.copy_from(data, temp_table)
            c.execute(f"""
                insert into {table}(namespace, name)
                    select namespace, name from {temp_table}
                    on conflict(namespace, name) do nothing;
                drop table {temp_table};
                commit
                """)
        except Exception as e:
            c.execute("rollback")
            raise
            
    @staticmethod
    def list(db, project=None):
        project_id = project.ID if isinstance(project, DBProject) else project
        c = db.cursor()
        table = DBFile.Table
        columns = DBFile.columns(as_text=True)
        files_columns = DBFile.columns(table, as_text=True)
        project_where = f" id = {project_id} " if project is not None else " true "
        c.execute(f"""
            select {columns}
                from {table}, projects
                where {project_where}
        """)
        return (DBFile.from_tuple(db, tup) for tup in cursor_iterator(c))

    @staticmethod
    def delete_many(db, specs):
        csv = [f"{namespace}:{name}"  for namespace, name in specs]
        data = io.StringIO("\n".join(csv))
        
        c = db.cursor()
        try:
            t = int(time.time()*1000)
            temp_table = f"files_temp_{t}"
            c.execute("begin")
            c.execute(f"creare temp table {temp_table} (spec text)")
            c.copy_from(data, temp_table)
            c.execute(f"""
                delete from {self.Table} 
                    where namespace || ':' || name in 
                        (select * from {temp_table});
                    on conflict (namespace, name) do nothing;       -- in case some other projects still use it
                drop table {temp_table};
                commit
                """)
        except Exception as e:
            c.execute("rollback")
            raise
    
class DBReplica(DBObject):
    Table = "replicas"
    Columns = ["namespace", "name", "rse", "path", "url", "preference", "available"]
    PK = ["namespace", "name", "rse"]
    
    def __init__(self, db, namespace, name, rse, path, url, preference=0, available=False):
        self.DB = db
        self.Namespace = namespace
        self.Name = name
        self.RSE = rse
        self.URL = url
        self.Path = path
        self.Preference = preference
        self.Available = available
        
    def did(self):
        return f"{self.Namespace}:{self.Name}"

    @staticmethod
    def list(db, namespace=None, name=None):
        c = db.cursor()
        wheres = " true "
        if namespace:   wheres += f" and namespace='{namespace}'"
        if name:        wheres += f" and name='{name}'"
        columns = DBReplica.columns(as_text=True)
        table = DBReplica.Table
        c.execute(f"""
            select {columns} from {table}
            where {wheres}
        """)
        return (DBReplica.from_tuple(db, tup) for tup in cursor_iterator(c))
        
    def as_jsonable(self):
        return dict(name=self.Name, namespace=self.Namespace, path=self.Path, 
            url=self.URL, rse=self.RSE,
            preference=self.Preference, available=self.Available)

    @staticmethod
    def create(db, namespace, name, rse, path, url, preference=0, available=False, error_if_exists=False):
        c = db.cursor()
        table = DBReplica.Table
        try:
            c.execute("begin")
            c.execute(f"""
                insert into {table}(namespace, name, rse, path, url, preference, available)
                    values(%s, %s, %s, %s, %s, %s, %s)
                    on conflict(namespace, name, rse)
                        do update set path=%s, url=%s, preference=%s, available=%s;
                commit
            """, (namespace, name, rse, path, url, preference, available,
                    path, url, preference, available)
            )
        except:
            c.execute("rollback")
            raise
            
        return DBReplica.get(db, namespace, name, rse)

    def save(self):
        table = self.Table
        c = self.DB.cursor()
        try:
            c.execute(f"""
                begin;
                update {table}
                     set path=%s, url=%s, preference=%s, available=%s
                     where namespace=%s and name=%s and rse=%s;
                commit
            """, (self.Path, self.URL, self.Preference, self.Available, self.Namespace, self.Name, self.RSE))
        except:
            c.execute("rollback")
            raise
        return self

    #
    # bulk operations
    #
    @staticmethod
    def remove_bulk(db, rse, dids):
        if not dids:    return
        c = db.cursor()
        table = DBReplica.Table
        try:
            c.execute(f"""
                begin;
                delete from {table}
                    where rse=%s and namespace || ':' || name = any(%s);
                commit
            """, (rse, dids))
        except:
            c.execute("rollback")
            raise

    @staticmethod
    def create_bulk(db, rse, preference, replicas):
        # replicas: {(namespace, name) -> {"path":.., "url":..}}
        # do not touch availability, update if exists

        csv = ['%s\t%s\t%s\t%s\t%s\t%s' % (namespace, name, rse, info["path"], info["url"], preference) 
            for (namespace, name), info in replicas.items()]
        #print("DBReplica.create_bulk: csv:", csv)
        data = io.StringIO("\n".join(csv))
        table = DBReplica.Table
        columns = DBReplica.columns(as_text=True)
        c = db.cursor()
        try:
            t = int(time.time()*1000)
            temp_table = f"file_replicas_temp_{t}"
            c.execute("begin")
            c.execute(f"create temp table {temp_table} (ns text, n text, r text, p text, u text, pr int)")
            c.copy_from(data, temp_table)
            c.execute(f"""
                insert into {table}({columns}) 
                    select t.ns, t.n, t.r, t.p, t.u, t.pr, false from {temp_table} t
                    on conflict (namespace, name, rse)
                        do nothing;
                drop table {temp_table};
                commit
                """)
        except Exception as e:
            c.execute("rollback")
            raise

    @staticmethod
    def update_availability_bulk(db, available, rse, dids):
        # dids is list of dids: ["namespace:name", ...]
        if not dids:    return
        table = DBReplica.Table
        val = "true" if available else "false"
        c = db.cursor()
        c.execute("begin")
        try:
            sql = f"""
                update {table}
                    set available=%s
                    where namespace || ':' || name = any(%s)
                        and rse = %s;
                commit
            """
            c.execute(sql, (val, dids, rse))
        except:
            c.execute("rollback")
            raise

class DBFileHandle(DBObject):

    Columns = ["project_id", "namespace", "name", "state", "worker_id", "attempts", "attributes"]
    PK = ["project_id", "namespace", "name"]
    Table = "file_handles"
    
    InitialState = ReadyState = "ready"
    ReservedState = "reserved"
    States = ["ready", "reserved", "done", "failed"]
    
    def __init__(self, db, project_id, namespace, name, state=None, worker_id=None, attempts=0, attributes={}):
        HasLogRecord.__init__(self, "file_handle_log", "file_handle_id")
        self.DB = db
        self.ProjectID = project_id
        self.Namespace = namespace
        self.Name = name
        self.State = state or self.InitialState
        self.WorkerID = worker_id
        self.Attempts = attempts
        self.Attributes = attributes.copy()
        self.File = None
        self.Replicas = None
        
    def id(self):
        return f"{self.ProjectID}:{self.Namespace}:{self.Name}"
        
    @staticmethod
    def unpack_id(id):
        project_id, namespace, name = id.split(":", 2)
        return int(project_id), namespace, name

    def did(self):
        return f"{self.Namespace}:{self.Name}"

    def get_file(self):
        if self.File is None:
            self.File = DBFile.get(self.DB, self.Namespace, self.Name)
        return self.File
        
    def replicas(self):
        if self.Replicas is None:
            self.Replicas = self.get_file().replicas()
        return self.Replicas
        
    def sorted_replicas(self):
        replicas = self.replicas().values()
        return sorted(replicas,
            key = lambda r: (
                0 if r.Available else 1,
                -r.Preference
            )
        )

    def as_jsonable(self, with_replicas=False):
        out = dict(
            project_id = self.ProjectID,
            namespace = self.Namespace,
            name = self.Name,
            state = self.State,
            worker_id = self.WorkerID,
            attempts = self.Attempts,
            attributes = self.Attributes
        )
        if with_replicas:
            out["replicas"] = {rse:r.as_jsonable() for rse, r in self.replicas().items()}
        return out
        
    def attributes_as_json(self):
        return json.dumps(self.Attributes, indent=4)
        
    @staticmethod
    def from_tuple(db, dbtup):
        #print("Handle.from_tuple: tuple:", dbtup)
        project_id, namespace, name, state, worker_id, attempts, attributes = dbtup
        h = DBFileHandle(db, project_id, namespace, name, state=state, worker_id=worker_id, attempts=attempts, attributes=attributes)
        return h
        
    @staticmethod
    def create(db, project_id, namespace, name, attributes={}):
        c = db.cursor()
        try:
            c.execute("begin")
            c.execute("""
                insert into file_handles(project_id, namespace, name, state, attempts, attributes)
                    values(%s, %s, %s, %s, 0, %s)
            """, (project_id, namespace, name, DBFileHandle.InitialState, json.dumps(attributes)))
            id = c.fetchone()[0]
            db.commit()
        except:
            c.execute("rollback")
            raise
        return DBFileHandle.get(db, project_id, namespace, name)
    
    @staticmethod
    def create_many(db, project_id, files):
        #
        # files: [ {"name":"...", "namespace":"...", "attributes":{}}]
        #
        files_csv = []
        parents_csv = []
        null = r"\N"
        for info in files:
            namespace = info["namespace"]
            name = info["name"]
            attributes = info.get("attributes", {})
            files_csv.append("%s\t%s\t%s\t%s\t%s" % (project_id, namespace, name, DBFileHandle.InitialState, json.dumps(attributes)))
        
        c = db.cursor()
        try:
            c.execute("begin")
            c.copy_from(io.StringIO("\n".join(files_csv)), "file_handles", 
                    columns = ["project_id", "namespace", "name", "state", "attributes"])
            c.execute("commit")
        except Exception as e:
            c.execute("rollback")
            raise
    
    @staticmethod
    def list(db, project_id=None, state=None, namespace=None, not_state=None, with_replicas=False):
        wheres = ["true"]
        if project_id: wheres.append(f"h.project_id={project_id}")
        if state:  
            if isinstance(state, (list, tuple)):
                wheres.append("h.state in (%s)" % ",".join(f"'{s}'" for s in state))
            else:
                wheres.append(f"h.state='{state}'")
        if not_state:       wheres.append(f"h.state!='{not_state}'")
        if namespace:       wheres.append(f"h.namespace='{namespace}'")
        wheres = " and ".join(wheres)
        c = db.cursor()
        h_columns = DBFileHandle.columns("h", as_text=True)
        r_columns = DBReplica.columns("r", as_text=True)
        h_n_columns = len(DBFileHandle.Columns)
        if with_replicas:
            sql = f"""
                select {h_columns}, {r_columns}
                    from file_handles h
                        left outer join replicas r on (r.name = h.name and r.namespace = h.namespace)
                        where {wheres}
                        order by h.namespace, h.name
            """
            #print("DBFileHandle.list: sql:", sql)
            c.execute(sql)
            h = None
            for tup in cursor_iterator(c):
                #print("DBFileHandle.list:", tup)
                h_tuple, r_tuple = tup[:h_n_columns], tup[h_n_columns:]
                if h is None:
                    h = DBFileHandle.from_tuple(db, h_tuple)
                h1 = DBFileHandle.from_tuple(db, h_tuple)
                if h1.Namespace != h.Namespace or h1.Name != h.Name:
                    if h:   
                        #print("    yield:", h)
                        yield h
                    h = h1
                if r_tuple[0] is not None:
                    r = DBReplica.from_tuple(db, r_tuple)
                    h.Replicas = h.Replicas or {}
                    h.Replicas[r.RSE] = r
            if h is not None:
                #print("    yield:", h)
                yield h

        else:
            sql = f"""
                select {h_columns}
                    from file_handles h
                        where {wheres}
            """
            c.execute(sql)
            yield from (DBFileHandle.from_tuple(db, tup) for tup in cursor_iterator(c))
                    
    def save(self):
        c = self.DB.cursor()
        try:
            c.execute("begin")
            c.execute("""
                update file_handles set state=%s, worker_id=%s, attempts=%s, attributes=%s
                    where project_id=%s and namespace=%s and name=%s
            """, (self.State, self.WorkerID, self.Attempts, json.dumps(self.Attributes), 
                    self.ProjectID, self.Namespace, self.Name
                )            
            )
            #print("DBFileHandle.save: attempts:", self.Attempts)
            self.DB.commit()
        except Exception as e:
            c.execute("rollback")
            raise
            
    @property
    def project(self):
        return DBProject.get(self.DB, self.ProjectID)
        
    @staticmethod
    def reserve_next_available(db, project_id, worker_id):
        # returns reserved handle or None
        c = db.cursor()
        columns = DBFileHandle.columns("h", as_text=True)
        sql = f"""
            update file_handles h
                    set state=%s, worker_id=%s, attempts = h.attempts + 1
                    where h.project_id=%s
                        and row(h.namespace, h.name) in
                        (       select hh.namespace, hh.name
                                        from file_handles hh, replicas r
                                        where hh.project_id=%s
                                                    and hh.state=%s
                                                    and hh.namespace = r.namespace 
                                                    and hh.name = r.name 
                                                    and r.available
                                        order by hh.attempts
                                        limit 1
                        )
                    returning {columns};
        """
        #print(sql)
        try:
            c.execute("begin")
            c.execute(sql, (DBFileHandle.ReservedState, worker_id, project_id, project_id, DBFileHandle.ReadyState))
        except:
            c.execute("rollback")
            raise
        tup = c.fetchone()
        if not tup:
            return None
        c.execute("commit")
        return DBFileHandle.from_tuple(db, tup)
        
    @staticmethod
    def available_handles(db, project_id, state):
        columns = self.columns("h", as_text=True)
        ncols = len(self.Columns)
        c = db.cursor()
        c.execute(f"""
            select {columns}, f.replicas
                from file_handles h
                    inner join files f on f.namespace = h.namespace and f.name = h.name
                where h.project_id=%s and h.state=%s and
                    f.replicas @? '$.* ? (@.available == true)';
                order h.attempts
        """, (project_id, DBFileHandle.InitialState))
        
        for tup in cursor_iterator(c):
            h = DBHandle.from_tuple(db, tup[:-1])
            h.Replicas = tup[-1]
            yield h

    def is_available(self):
        return any(r.Available for r in self.replicas().values())
        
    def done(self):
        self.State = "done"
        self.WorkerID = None
        self.save()
        
    def failed(self, retry=True):
        self.State = self.ReadyState if retry else "failed"
        self.WorkerID = None
        self.save()
        
    def is_active(self):
        return self.State not in ("done", "failed")
        



from data_dispatcher.api import DataDispatcherClient
from metacat.webapi import MetaCatClient 
import sys, getopt, os, json
from .ui_project import create_project, show_project, list_projects
from .ui_handle import show_handle
from .ui_lib import pretty_json
from .ui_file import show_file, replica_available, replica_unavailable
import time

server_url = os.environ.get("DATA_DISPATCHER_URL")
if not server_url:
    print("Data Dispatcher Server URL undefined. Use DATA_DISPATCHER_URL environment variable")
    sys.exit(2)

my_name = sys.argv[0]

    
Usage = f"""
Usage:

dd <commnd> <subcommand> <options> <args>

Commands:

    worker [-n]                                         - print my worker_id. -n generates new one.
                                                          worker id will be saved in <CWD>/.worker_id
    create project ...
    show project ...
    list projects ...
    
    replica available ...
    replica unavailable ...

    show file [-j] <namespace>:<name>

    list handles [-s <status>] <project_id>                - list file handles, -s filters handles by status
    show handle [-j] <project_id> <namespace>:<name>       - show file handle, -j - as JSON

    next [-j] [-w <timeout>|-W] <project_id>               - get next available file, 
                                                             -j - as JSON
                                                             -w - wait for next file until "timeout" seconds
                                                             -W - wait for next file until the project finishes

    done <project_id> <namespace>:<name>                   - mark the file as successfully processed
    failed [-f] <project_id> <namespace>:<name>            - mark the file as failed, -f means "final", no retries

    delete project <project_id>
    
    login x509 <user> <cert> <key>
    login password <user>
"""

def main():
    client = DataDispatcherClient(server_url)

    if len(sys.argv[1:]) < 1 or "-\?" in sys.argv[1:] or "--help" in sys.argv[1:]:
        print(Usage)
        sys.exit(2)
    
    command = sys.argv[1]
    rest = sys.argv[2:]

    if command == "worker":
        if rest and rest[0] == "-n":
            worker_id = client.new_worker_id()
        else:
            worker_id = client.WorkerID
        if not worker_id:
            print("worker id unknown")
            sys.exit(1)
        print(worker_id)
        
    elif command == "create":
        subcommand, rest = rest[0], rest[1:]
        if subcommand != "project":
            print(Usage)
            sys.exit(2)
        create_project(client, rest)
    
    elif command == "show":
        subcommand, rest = rest[0], rest[1:]
        if subcommand not in ("file", "project", "handle"):
            print(Usage)
            sys.exit(2)
        if subcommand == "project":
            show_project(client, rest)
        elif subcommand == "handle":
            show_handle(client, rest)
        elif subcommand == "file":
            show_file(client, rest)
        else:
            print(Usage)
            sys.exit(2)
        
    elif command == "replica":
        subcommand, rest = rest[0], rest[1:]
        if subcommand == "available":
            replica_available(client, rest)
        elif subcommand == "unavailable":
            replica_unavailable(client, rest)
        else:
            print(Usage)
            sys.exit(2)

    elif command == "list":
        if not rest:
            print(Usage)
            sys.exit(2)
        subcommand, rest = rest[0], rest[1:]
        if subcommand == "projects":
            list_projects(client, rest)
        elif subcommand == "handles":
            list_handles(client, rest)
        else:
            print(Usage)
            sys.exit(2)
        
    elif command == "delete":
        subcommand, rest = rest[0], rest[1:]
        if subcommand == "project":
            project_id = rest[0]
            client.delete_project(project_id)
        else:
            print(Usage)
            sys.exit(2)
        

    elif command == "next":
        opts, args = getopt.getopt(rest, "jw:W")
        opts = dict(opts)
        if not args:
            print(Usage)
            sys.exit(2)
        project_id = args[0]
        as_json = "-j" in opts
        timeout = int(opts.get("-w", 0))
        if "-W" in opts:
            timeout = None
        wait = "-w" in opts or "-W" in opts
        done = False
        t0 = time.time()
        while not done:
            info = client.next_file(project_id)
            if info:
                if as_json:
                    print(pretty_json(info))
                else:
                    print("%s:%s" % (info["namespace"], info["name"]))
                sys.exit(0)
            if wait and (timeout is None or time.time() < t0 + timeout):
                project_info = client.get_project(project_id)
                if not project_info["active"]:
                    sys.exit(10)    # project finished
                time.sleep(5)
            else:
                sys.exit(11)        # timeout

    elif command == "done":
        if len(rest) != 2:
            print(Usage)
            sys.exit(2)
        project_id, did = rest
        client.file_done(project_id, did)
    
    elif command == "failed":
        opts, args = getopt.getopt(rest, "f")
        opts = dict(opts)
        if len(args) != 2:
            print(Usage)
            sys.exit(2)
        project_id, did = args
        client.file_failed(project_id, did, retry = not "-f" in opts)
    
    elif command == "login":
        if len(rest) < 2:
            print(Usage)
            sys.exit(2)
        subcommand, rest = rest[0], rest[1:]
        if subcommand == "x509":
            username, cert, key = rest
            user, expiration = client.login_x509(username, cert, key)
        elif subcommand == "password":
            import getpass
            username = rest[0]
            password = getpass.getpass("Password:")
            user, expiration = client.login_password(username, password)
        else:
            print(f"Unknown authentication mechanism {subcommand}\n")
            print(Usage)
            sys.exit(2)

        print ("User:   ", user)
        print ("Expires:", time.ctime(expiration))
    
    else:
        print(Usage)
        sys.exit(2)

if __name__ == "__main__":
    main()
        
        
    
            
            
        
        
        
        
        



import requests, uuid, json, urllib.parse, os
from metacat.auth import TokenLib, TokenAuthClientMixin

import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class ServerError(Exception):
    
    def __init__(self, url, status_code, message, body=""):
        self.URL = url
        self.StatusCode = status_code
        self.Message = message
        self.Body = to_str(body)
        
    def __str__(self):
        msg = f"DataDispatcherServer error:\n  URL: {self.URL}\n  HTTP status code: {self.StatusCode}\n  Message: {self.Message}"
        if self.Body:
            msg += "\nMessage from the server:\n"+self.Body+"\n"
        return msg

class APIError(ServerError):
    
    def __init__(self, url, status_code, body):
        ServerError.__init__(self, url, status_code, "", body)
    
    def json(self):
        #print("WebAPIError.json: body:", self.Body)
        return json.loads(self.Body)


def to_bytes(x):
    if not isinstance(x, bytes):
        x = x.encode("utf-8")
    return x
    
def to_str(x):
    if isinstance(x, bytes):
        x = x.decode("utf-8")
    return x
    
class HTTPClient(object):

    def __init__(self, server_url, token=None):
        self.ServerURL = server_url
        self.Token = token

    def get(self, uri_suffix):
        if not uri_suffix.startswith("/"):  uri_suffix = "/"+uri_suffix
        url = "%s%s" % (self.ServerURL, uri_suffix)
        headers = {}
        if self.Token is not None:
            headers["X-Authentication-Token"] = self.Token.encode()
        response = requests.get(url, headers =headers)
        #print("response status code:", response.status_code)
        if response.status_code != 200:
            raise APIError(url, response.status_code, response.text)
        if response.headers.get("Content-Type", "").startswith("text/json"):
            data = json.loads(response.text)
        else:
            data = response.text
        return data
        
    def post(self, uri_suffix, data):
        #print("post_json: data:", type(data), data)
        
        if not uri_suffix.startswith("/"):  uri_suffix = "/"+uri_suffix
        
        if data is None or isinstance(data, (dict, list)):
            data = json.dumps(data)
        else:
            data = to_bytes(data)
        #print("post_json: data:", type(data), data)
            
        url = "%s%s" % (self.ServerURL, uri_suffix)
        
        headers = {}
        if self.Token is not None:
            headers["X-Authentication-Token"] = self.Token.encode()
        #print("HTTPClient.post_json: url:", url)
        #print("HTTPClient.post_json: data:", data)
        
        response = requests.post(url, data = data, headers = headers)
        if response.status_code != 200:
            raise APIError(url, response.status_code, response.text)
        #print("response.text:", response.text)
        if response.headers.get("Content-Type", "").startswith("text/json"):
            data = json.loads(response.text)
        else:
            data = response.text
        return data
        
class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
    
    DefaultWorkerIDFile = ".data_dispatcher_worker_id"
    
    def __init__(self, server_url, worker_id=None, worker_id_file=None, token = None, token_file = None, auth_server_url=None):
        server_url = server_url or os.environ.get("DATA_DISPATCHER_URL")
        auth_server_url = auth_server_url or os.environ.get("DATA_DISPATCHER_AUTH_URL")
        TokenAuthClientMixin.__init__(self, server_url, token, token_file, auth_url=auth_server_url)

        #print("DataDispatcherClient: url:", server_url)

        worker_id_file = worker_id_file or self.DefaultWorkerIDFile
        if worker_id is None:
            if worker_id_file:
                try:    worker_id = open(worker_id_file, "r").read().strip()
                except: pass
        if worker_id is None:
            worker_id = self.gen_worker_id()
            if worker_id_file:
                open(worker_id_file, "w").write(worker_id)
        self.WorkerID = worker_id
        
        HTTPClient.__init__(self, server_url, self.token())
            
        #print("DataDispatcherClient: token:", token, "  user:", token["user"])
        
    def gen_worker_id(self):
        u = uuid.uuid4().bytes
        n = len(u)//4
        wid = [0]*n
        for i in range(n):
            for j in range(0, n, 4):
                wid[i] ^= u[j+i]
        out = bytes(wid).hex()
        #print("gen_worker_id:", out)
        return out
        
    def new_worker_id(self, worker_id_file=None):
        worker_id_file = worker_id_file or self.DefaultWorkerIDFile
        worker_id = self.gen_worker_id()
        #print("generated worker_id:", worker_id)
        open(worker_id_file, "w").write(worker_id)
        self.WorkerID = worker_id
        return worker_id

    #
    # projects
    #
    FileFields = ("namespace","name","attributes")
    def create_project(self, files, attributes={}):
        cleaned = []
        for info in files:
            cleaned.append({name:info.get(name) for name in self.FileFields})
        return self.post("create_project", json.dumps(
            {   
                "files":        cleaned,
                "attributes":   attributes
            }
            )
        )

    def delete_project(self, project_id):
        return self.get(f"delete_project?project_id={project_id}")
        
    def get_project(self, project_id, with_files=True, with_replicas=False):
        with_files = "yes" if with_files else "no"
        with_replicas = "yes" if with_replicas else "no"
        uri = f"project?project_id={project_id}&with_files={with_files}&with_replicas={with_replicas}"
        return self.get(uri)
    
    def list_projects(self, owner=None, state=None, not_state=None, attributes=None, with_files=True, with_replicas=False):
        suffix = "projects"
        args = []
        if owner: args.append(f"owner={owner}")
        if state: args.append(f"state={state}")
        if not_state: args.append(f"not_state={not_state}")
        args.append("with_handles=" + ("yes" if with_files else "no"))
        args.append("with_replicas=" + ("yes" if with_replicas else "no"))
        if attributes: args.append("attributes="+urllib.parse.quote(json.dumps(attributes)))
        args = "?" + "&".join(args) if args else ""
        return self.get(f"projects{args}")

    def next_file(self, project_id):
        if self.WorkerID is None:
            raise ValueError("DataDispatcherClient must be initialized with Worker ID")
        return self.get(f"next_file?project_id={project_id}&worker_id={self.WorkerID}")
        
    #
    # File handles
    #
    
    def list_handles(self, project_id=None, state=None, not_state=None, rse=None):
        args = []
        if rse: args.append(f"rse={rse}")
        if project_id: args.append(f"project_id={project_id}")
        if state: args.append(f"state={state}")
        if not_state: args.append(f"not_state={not_state}")
        args = "?" + "&".join(args) if args else ""
        return self.get(f"handles{args}")

    def find_handle(self, project_id, name=None, file_id=None, namespace=None):
        if file_id:
            suffix = f"file_handle?project_id={project_id}&file_id={file_id}"
        else:
            suffix = f"file_handle?project_id={project_id}&name={name}"
            if namespace:
                suffix += f"&namespace={namespace}"
        return self.get(suffix)

    def get_handle(self, handle_id):
        return self.get(f"file_handle?handle_id={handle_id}")

    def get_file(self, namespace, name):
        return self.get(f"file?namespace={namespace}&name={name}")

    def replica_available(self, namespace, name, rse, path=None, preference=0, url=None):
        data = {
            "path": path,
            "url": url,
            "preference": preference
        }
        suffix = f"replica_available?&rse={rse}&namespace={namespace}&name={name}"
        return self.post(suffix, data)

    def replica_unavailable(self, namespace, name, rse):
        suffix = f"replica_unavailable?&rse={rse}&namespace={namespace}&name={name}"
        return self.get(suffix)

    def file_done(self, project_id, did):
        handle_id = f"{project_id}:{did}"
        return self.get(f"release?handle_id={handle_id}&failed=no")

    def file_failed(self, project_id, did, retry=True):
        handle_id = f"{project_id}:{did}"
        retry = "yes" if retry else "no"
        return self.get(f"release?handle_id={handle_id}&failed=yes&retry={retry}")
        
    
        
        
        



import getopt, json, time
from metacat.webapi import MetaCatClient
from .ui_lib import pretty_json, parse_attrs, print_handles


create_project_help = """
create project [options] [inline MQL query]
        -q <file with MQL query>                    - read MQL query from file instead
        -c <name>[,<name>...]                       - copy metadata attributes from query results, use with -q or inline query

        -l (-|<file with file list>)                - read "namespace:name [attr=value ...]" lines from the file 
        -j (-|<file with file list>)                - read JSON file with file list {"namespace":...,"name"..., "attributes":{...}}

        -A @<file.json>                             - JSON file with project attributes
        -A "name=value name=value ..."              - project attributes
        -a @<file.json>                             - JSON file with common file attributes
        -a "name=value name=value ..."              - common file attributes

        -p (json|pprint|id)                         - print created project info as JSON, pprint or just project id (default)
"""


def create_project(client, rest):
    if not rest or "help" in rest:
        print(create_project_help)
        return
    opts, args = getopt.getopt(rest, "l:q:p:a:A:c:j:")
    opts = dict(opts)
    query = " ".join(args).strip()

    common_attrs = {}
    if "-a" in opts:
        attr_src = opts["-a"]
        if attr_src[0] == '@':
            common_attrs = json.load(open(attr_src[1:], "r"))
        else:
            common_attrs = parse_attrs(attr_src)
 
    project_attrs = {}
    if "-A" in opts:
        attr_src = opts["-A"]
        if attr_src[0] == '@':
            project_attrs = json.load(open(attr_src[1:], "r"))
        else:
            project_attrs = parse_attrs(attr_src)

    if not query and "-q" in opts:
            query = open(opts["-q"], "r").read().strip()
        
    if query:
        files = MetaCatClient().query(query, with_metadata = "-c" in opts)
        for info in files:
            info["attributes"] = common_attrs.copy()
        #
        # copy file attributes from metacat
        #
        
        if "-c" in opts:
            fields = opts["-c"].split(",")
            for info in files:
                #print(info["metadata"])
                attrs = {k:info["metadata"].get(k) for k in fields}
                info["attributes"].update(attrs)    
        
    elif "-j" in opts:
        inp = opts["-j"]
        if inp == "-":
            inp = sys.stdin
        else:
            inp = open(inp, "r")
        files = json.load(inp)
        if common_attrs:
            for info in files:
                attrs = info.get("attributes", {}).copy()
                attrs.update(common_attrs)
                info["attributes"] = attrs

    elif "-l" in opts:
        inp = opts["-l"]
        if inp == "-":
            inp = sys.stdin
        else:
            inp = open(inp, "r")
        files = []
        for line in inp:
            line = line.strip()
            if line:
                did, rest = (tuple(line.split(None, 1)) + (None,))[:2]
                namespace, name = did.split(":", 1)
                attrs = common_attrs.copy()
                if rest:
                    attrs.update(parse_attrs(rest))
                files.append({"namespace":namespace, "name":name, "attributes":attrs})

    #print("files:", files)
    info = client.create_project(files, attributes=project_attrs)
    printout = opts.get("-p", "id")
    if printout == "json":
        print(pretty_json(info))
    elif printout == "pprint":
        pprint.pprint(info)
    else:
        print(info["project_id"])

show_project_help = """
show project [options] <project_id>             - show project info (-j show as JSON)
    -a                                          - show project attributes only
    -r                                          - show replicas information
    -j                                          - show as JSON
    -f [active|ready|available|all|reserved|failed|done]    - list files (namespace:name) only
                                                   all       - all files, including done and failed
                                                   active    - all except done and failed
                                                   ready     - ready files only
                                                   available - available files only
                                                   reserved  - reserved files only
                                                   failed    - failed files only
                                                   done      - done files only
"""


def show_project(client, rest):
    if not rest or "help" in rest:
        print(show_project_help)
        return
    opts, args = getopt.getopt(rest, "jarf:")
    opts = dict(opts)
    project_id = args[0]
    info = client.get_project(project_id, with_files=True, with_replicas=True)
    if "-a" in opts:
        attrs = info.get("attributes", {})
        if "-j" in opts:
            print(pretty_json(attrs))
        else:
            for name, value in sorted(attrs.items()):
                print(f"{name} {value}")
    else:
        if "-j" in opts:
            print(pretty_json(info))
        elif "-f" in opts:
            filter_state = opts["-f"]
            for h in info["file_handles"]:
                did = h["namespace"] + ":" + h["name"]
                state = h["state"]
                rlist = h["replicas"].values()
                available = state == "ready" and len([r for r in rlist if r["available"]]) > 0
                if filter_state == "all" or \
                            filter_state in ("done", "ready", "failed", "reserved") and state == filter_state or \
                            filter_state == "available" and available or \
                            filter_state == "active" and not state in ("done", "failed"):
                    print(did)
        else:
            print("Project ID:         ", project_id)
            print("Status:             ", info["state"])
            print("Project Attributes: ")
            for k, v in info.get("attributes", {}).items():
                if not isinstance(v, (int, float, str, bool)):
                    v = json.dumps(v)
                print("  %-15s = %s" % (k, v))
            print("Files:              ")
            print_replicas = "-r" in opts
            if "file_handles" in info:
                print_handles(info["file_handles"], print_replicas)
    
list_projects_help = """
list projects     
        -j                                              - JSON output
        -u <owner>                                      - filter by owner
        -s <state>                                      - filter by state
        -a "name=value name=value ..."                  - filter by attributes
"""

def list_projects(client, rest):
    if "help" in rest or "--help" in rest or "-?" in rest:
        print(list_projects_help)
        return
    opts, args = getopt.getopt(rest, "s:a:ju:")
    opts = dict(opts)
    state = opts.get("-s")
    attributes = None
    if "-a" in opts:
        attributes = parse_attrs(opts["-a"])
    owner = opts.get("-u")
    lst = client.list_projects(state=state, attributes=attributes, owner=owner, with_files=True, with_replicas=False)
    if "-j" in opts:
        print(pretty_json(list(lst)))
    else:
        print("%-15s %-15s %-19s %-15s %17s" % ("project id", "owner", "created", "state", "done/failed/files"))
        print("%s %s %s %s %s" % ("-"*15, "-"*15, "-"*19, "-"*15, "-"*17))
        for prj in lst:
            ct = time.localtime(prj["created_timestamp"])
            ct = time.strftime("%Y-%m-%d %H:%M:%S", ct)
            
            nfiles = len(prj["file_handles"])
            ready_files = reserved_files = failed_files = done_files = available_files = 0
            for h in prj["file_handles"]:
                if h["state"] == "ready":
                    ready_files += 1
                elif h["state"] == "done":
                    done_files += 1
                elif h["state"] == "failed":
                    failed_files += 1
            counts = "%d/%d/%d" % (done_files, failed_files, nfiles)
            print("%-15s %-15s %19s %15s %17s" % (prj["project_id"], prj["owner"], ct, prj["state"], counts))
        print("%s %s %s %s %s" % ("-"*15, "-"*15, "-"*19, "-"*15, "-"*17))
    



import getopt, sys
from .ui_lib import to_did, from_did, pretty_json

Usage = """
replica available [options] <rse> <file_namespace:file_name>        - add/modify available file replica
    -u <url>
    -p <path>
    -n <preference>     - default: 0
    -j                  - print final file info as JSON

replica unavailable <rse> <file_namespace:file_name>                - mark file replica unavailable

show file [-j] <file_namespace:file_name>                   - show file info (-j = as JSON)
"""

def replica_available(client, rest):
        opts, args = getopt.getopt(rest, "p:n:u:j")
        opts = dict(opts)
        if len(args) != 2:
            print(Usage)
            sys.exit(2)
        rse, did = args
        namespace, name = from_did(did)
        path = opts.get("-p")
        url = opts.get("-u")
        preference = int(opts.get("-n", 0))
        data = client.replica_available(namespace, name, rse, path=path, url=url, preference=preference)
        if "-j" in opts:
            print(pretty_json(data))

def replica_unavailable(client, rest):
        opts, args = getopt.getopt(rest, "j")
        opts = dict(opts)
        if len(args) != 2:
            print(Usage)
            sys.exit(2)
        rse, did = args
        namespace, name = from_did(did)
        data = client.replica_unavailable(namespace, name, rse)
        if "-j" in opts:
            print(pretty_json(data))
            
def show_file(client, rest):
        opts, args = getopt.getopt(rest, "j")
        did = args[0]
        namespace, name = from_did(did)
        data = client.get_file(namespace, name)
        if "-j" in opts:
            print(pretty_json(data))
        else:
            replicas = data["replicas"].items()
            replicas = sorted(replicas, key=lambda info: (not info[1].get("available"), info[0]))
            print("Namespace: ", namespace)
            print("Name:      ", name)
            print("Replicas:  ", len(replicas))
            for rse, info in replicas:
                print("%1s %-25s %s" % ("A" if info["available"] else "U", rse, info.get("url") or ""))
                if info.get("path"):
                    print("%27s %s" % ("", info["path"]))



import traceback, sys, time
from pythreader import LogFile, LogStream

class Logger(object):

    def __init__(self, log_output, debug_enabled=False, debug_out=None, error_out=None):
        self.LogOut = self.make_log(log_output, sys.stdout)
        self.ErrorOut = self.make_log(error_out, sys.stderr)
        self.DebugOut = None if not debug_enabled else self.make_log(debug_out, sys.stdout)
        self.DebugEnabled = debug_enabled
        
    def make_log(self, output, dash_stream, **params):
        if output is None:  
            return None
        elif isinstance(output, (LogFile, LogStream)):
            return output
        elif output == "-":
            return LogStream(dash_stream)
        elif output is sys.stderr or output is sys.stdout:
            return LogStream(output)
        else:
            out = LogFile(output, **params)
            out.start()
            return out
            
    def log(self, who, *parts, sep=" "):
        if self.LogOut is not None:
            self.LogOut.log("%s: %s" % (who, sep.join([str(p) for p in parts])))
            
    def debug(self, who, *parts, sep=" "):
        if self.DebugEnabled and self.DebugOut is not None:
            self.DebugOut.log("%s [DEBUG]: %s" % (who, sep.join([str(p) for p in parts])))

    def error(self, who, *parts, sep=" "):
        if self.ErrorOut is not None:
            self.ErrorOut.log("%s [ERROR]: %s" % (who, sep.join([str(p) for p in parts])))
            
class Logged(object):

    def __init__(self, name, logger, debug=True):
        self.LogName = name
        self.Logger = logger
        self.Debug = debug
        
    def debug(self, *params):
        if self.Debug:
            self.Logger.debug(self.LogName, *params)

    def log(self, *params):
        self.Logger.log(self.LogName, *params)

    def error(self, *params):
        self.Logger.error(self.LogName, *params)




import traceback, sys, time
from pythreader import LogFile, LogStream

LogOut = ErrorOut = DebugOut = None
DebugEnabled = False

class Logged(object):

    def __init__(self, name=None, debug=True):
        self.LogName = name or self.__class__.__name__
        self.Debug = debug
        
    def debug(self, *params, sep=" "):
        if self.Debug:
            LogOut.log("%s [DEBUG]: %s" % (self.LogName, sep.join([str(p) for p in params])))

    def log(self, *params, sep=" "):
        LogOut.log("%s: %s" % (self.LogName, sep.join([str(p) for p in params])))

    def error(self, *params, sep=" "):
        LogOut.log("%s [ERROR]: %s" % (self.LogName, sep.join([str(p) for p in params])))

def make_log(output, dash_stream, **params):
    if output is None:  
        return None
    elif isinstance(output, (LogFile, LogStream)):
        return output
    elif output == "-":
        return LogStream(dash_stream)
    elif output is sys.stderr or output is sys.stdout:
        return LogStream(output)
    else:
        out = LogFile(output, **params)
        out.start()
        return out
            
def init_logger(log_output, debug_enabled=False, debug_out=None, error_out=None):
    global LogOut, ErrorOut, DebugOut, DebugEnabled
    LogOut = make_log(log_output, sys.stdout)
    ErrorOut = make_log(error_out, sys.stderr)
    DebugOut = None if not debug_enabled else make_log(debug_out, sys.stdout)
    DebugEnabled = debug_enabled
    
